@@ -12,14 +12,15 @@
 
 import json
 import re
+from unittest import mock
 
-import mock
 from testtools import matchers
 
 from keystoneauth1 import adapter
 from keystoneauth1 import discover
 from keystoneauth1 import exceptions
 from keystoneauth1 import fixture
+from keystoneauth1 import http_basic
 from keystoneauth1 import noauth
 from keystoneauth1 import session
 from keystoneauth1.tests.unit import utils
@@ -560,6 +561,102 @@ class VersionDataTests(utils.TestCase):
 
         self.assertTrue(mock.called_once)
 
+    def test_version_data_legacy_ironic_no_override(self):
+        """Validate detection of legacy Ironic microversion ranges."""
+        ironic_url = 'https://bare-metal.example.com/v1/'
+        self.requests_mock.get(
+            ironic_url, status_code=200,
+            json={
+                'id': 'v1',
+                'links': [{
+                    "href": ironic_url,
+                    "rel": "self"}]},
+            headers={
+                'X-OpenStack-Ironic-API-Minimum-Version': '1.3',
+                'X-OpenStack-Ironic-API-Maximum-Version': '1.21',
+            })
+
+        plugin = noauth.NoAuth()
+        a = adapter.Adapter(
+            self.session,
+            auth=plugin,
+            service_type='baremetal')
+
+        self.assertIsNone(a.get_api_major_version())
+
+    def test_version_data_ironic_microversions(self):
+        """Validate detection of Ironic microversion ranges."""
+        ironic_url = 'https://bare-metal.example.com/v1/'
+        self.requests_mock.get(
+            ironic_url, status_code=200,
+            json={
+                'id': 'v1',
+                'version': {
+                    'id': 'v1',
+                    'links': [{
+                        "href": ironic_url,
+                        "rel": "self"}],
+                    'version': '1.40',
+                    'min_version': '1.10',
+                    'status': 'CURRENT',
+                },
+                'links': [{
+                    "href": ironic_url,
+                    "rel": "self"}],
+            },
+            # Keep headers so we can verify that body trumps headers
+            headers={
+                'X-OpenStack-Ironic-API-Minimum-Version': '1.3',
+                'X-OpenStack-Ironic-API-Maximum-Version': '1.21',
+            })
+
+        self.assertEqual(
+            [
+                {
+                    'collection': None,
+                    'version': (1, 0),
+                    'url': ironic_url,
+                    'status': discover.Status.CURRENT,
+                    'raw_status': discover.Status.CURRENT,
+                    'min_microversion': (1, 10),
+                    'max_microversion': (1, 40),
+                    'next_min_version': None,
+                    'not_before': None,
+                },
+            ],
+            discover.Discover(self.session, ironic_url).version_data())
+
+    def test_version_data_legacy_ironic_microversions(self):
+        """Validate detection of legacy Ironic microversion ranges."""
+        ironic_url = 'https://bare-metal.example.com/v1/'
+        self.requests_mock.get(
+            ironic_url, status_code=200,
+            json={
+                'id': 'v1',
+                'links': [{
+                    "href": ironic_url,
+                    "rel": "self"}]},
+            headers={
+                'X-OpenStack-Ironic-API-Minimum-Version': '1.3',
+                'X-OpenStack-Ironic-API-Maximum-Version': '1.21',
+            })
+
+        self.assertEqual(
+            [
+                {
+                    'collection': None,
+                    'version': (1, 0),
+                    'url': ironic_url,
+                    'status': discover.Status.CURRENT,
+                    'raw_status': discover.Status.CURRENT,
+                    'min_microversion': (1, 3),
+                    'max_microversion': (1, 21),
+                    'next_min_version': None,
+                    'not_before': None,
+                },
+            ],
+            discover.Discover(self.session, ironic_url).version_data())
+
     def test_version_data_microversions(self):
         """Validate [min_|max_]version conversion to {min|max}_microversion."""
         def setup_mock(versions_in):
@@ -651,24 +748,81 @@ class VersionDataTests(utils.TestCase):
 
     def test_endpoint_data_noauth_discover(self):
         mock = self.requests_mock.get(
+            BASE_URL, status_code=200, json=V3_VERSION_LIST)
+        self.requests_mock.get(
             V3_URL, status_code=200, json=V3_VERSION_ENTRY)
+
+        plugin = noauth.NoAuth(endpoint=BASE_URL)
+        data = plugin.get_endpoint_data(self.session)
+
+        self.assertEqual(data.api_version, (3, 0))
+        self.assertEqual(data.url, V3_URL)
+        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
+        self.assertEqual(plugin.get_endpoint(self.session), BASE_URL)
+
+        self.assertTrue(mock.called_once)
+
+    def test_endpoint_data_noauth_versioned_discover(self):
+        self.requests_mock.get(
+            BASE_URL, status_code=200, json=V3_VERSION_LIST)
+        self.requests_mock.get(
+            V3_URL, status_code=200, json=V3_VERSION_ENTRY)
+
+        plugin = noauth.NoAuth(endpoint=V3_URL)
+        data = plugin.get_endpoint_data(self.session)
+
+        self.assertEqual(data.api_version, (3, 0))
+        self.assertEqual(data.url, V3_URL)
+        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
+        self.assertEqual(plugin.get_endpoint(self.session), V3_URL)
+
+    def test_endpoint_data_noauth_no_discover(self):
+        plugin = noauth.NoAuth(endpoint=V3_URL)
+        data = plugin.get_endpoint_data(
+            self.session, discover_versions=False)
+
+        self.assertEqual(data.api_version, (3, 0))
+        self.assertEqual(data.url, V3_URL)
+        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
+        self.assertEqual(plugin.get_endpoint(self.session), V3_URL)
+
+    def test_endpoint_data_noauth_override_no_discover(self):
         plugin = noauth.NoAuth()
-        data = plugin.get_endpoint_data(self.session, endpoint_override=V3_URL)
+        data = plugin.get_endpoint_data(
+            self.session, endpoint_override=V3_URL, discover_versions=False)
 
         self.assertEqual(data.api_version, (3, 0))
         self.assertEqual(data.url, V3_URL)
-        self.assertEqual(
-            plugin.get_api_major_version(
-                self.session, endpoint_override=V3_URL),
-            (3, 0))
         self.assertEqual(
             plugin.get_endpoint(self.session, endpoint_override=V3_URL),
             V3_URL)
 
-        self.assertTrue(mock.called_once)
+    def test_endpoint_data_http_basic_discover(self):
+        self.requests_mock.get(
+            BASE_URL, status_code=200, json=V3_VERSION_LIST)
+        self.requests_mock.get(
+            V3_URL, status_code=200, json=V3_VERSION_ENTRY)
 
-    def test_endpoint_data_noauth_no_discover(self):
-        plugin = noauth.NoAuth()
+        plugin = http_basic.HTTPBasicAuth(endpoint=V3_URL)
+        data = plugin.get_endpoint_data(self.session)
+
+        self.assertEqual(data.api_version, (3, 0))
+        self.assertEqual(data.url, V3_URL)
+        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
+        self.assertEqual(plugin.get_endpoint(self.session), V3_URL)
+
+    def test_endpoint_data_http_basic_no_discover(self):
+        plugin = http_basic.HTTPBasicAuth(endpoint=V3_URL)
+        data = plugin.get_endpoint_data(
+            self.session, discover_versions=False)
+
+        self.assertEqual(data.api_version, (3, 0))
+        self.assertEqual(data.url, V3_URL)
+        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
+        self.assertEqual(plugin.get_endpoint(self.session), V3_URL)
+
+    def test_endpoint_data_http_basic_override_no_discover(self):
+        plugin = http_basic.HTTPBasicAuth()
         data = plugin.get_endpoint_data(
             self.session, endpoint_override=V3_URL, discover_versions=False)
 
@@ -683,6 +837,22 @@ class VersionDataTests(utils.TestCase):
             V3_URL)
 
     def test_endpoint_data_noauth_adapter(self):
+        self.requests_mock.get(
+            BASE_URL, status_code=200, json=V3_VERSION_LIST)
+        self.requests_mock.get(
+            V3_URL, status_code=200, json=V3_VERSION_ENTRY)
+
+        client = adapter.Adapter(
+            session.Session(noauth.NoAuth()),
+            endpoint_override=BASE_URL)
+        data = client.get_endpoint_data()
+
+        self.assertEqual(data.api_version, (3, 0))
+        self.assertEqual(data.url, V3_URL)
+        self.assertEqual(client.get_api_major_version(), (3, 0))
+        self.assertEqual(client.get_endpoint(), BASE_URL)
+
+    def test_endpoint_data_noauth_versioned_adapter(self):
         mock = self.requests_mock.get(
             V3_URL, status_code=200, json=V3_VERSION_ENTRY)
 
@@ -1145,6 +1315,29 @@ class EndpointDataTests(utils.TestCase):
             [mock.call('sess', url, cache='cache', authenticated=False)
              for url in ('url1', 'url2', 'url3')])
 
+    def test_run_discovery_auth(self):
+        url = 'https://example.com'
+        headers = {'Accept': 'application/json'}
+
+        session = mock.Mock()
+        session.get.side_effect = [
+            exceptions.Unauthorized('unauthorized'),
+            # Throw a different exception the second time so that we can
+            # catch it in the test and verify the retry.
+            exceptions.BadRequest('bad request'),
+        ]
+
+        try:
+            discover.get_version_data(session, url)
+        except exceptions.BadRequest:
+            pass
+        # Only one call with 'url'
+        self.assertEqual(2, session.get.call_count)
+        session.get.assert_has_calls([
+            mock.call(url, headers=headers, authenticated=None),
+            mock.call(url, headers=headers, authenticated=True),
+        ])
+
     def test_endpoint_data_str(self):
         """Validate EndpointData.__str__."""
         # Populate a few fields to make sure they come through.
@@ -1161,3 +1354,8 @@ class EndpointDataTests(utils.TestCase):
         self.assertEqual(exp, str(epd))
         # Works with implicit stringification
         self.assertEqual(exp, "%s" % epd)
+
+    def test_project_id_int_fallback(self):
+        bad_url = "https://compute.example.com/v2/123456"
+        epd = discover.EndpointData(catalog_url=bad_url)
+        self.assertEqual((2, 0), epd.api_version)
